import os
import random
import threading
import logging
import socket
import datetime
from collections import defaultdict
from time import sleep

from pymongo import MongoClient
from scrapy import Request
from scrapy.exceptions import IgnoreRequest, DropItem, CloseSpider
from scrapy.http import Request, HtmlResponse
from scrapy.spiders import CrawlSpider, Rule
from scrapy.utils.spider import iterate_spider_output
from scrapy.linkextractors import LinkExtractor
from scrapy_projects.itzy_bitzy_project.items import URLItem
from random import shuffle
from hashlib import sha256

class BaseBitsySpider(CrawlSpider):
    name = "bitsy"
    mongo_uri = "mongodb://localhost"
    mongo_db = "webcoin"


    path_1_zeros_needed = 10

    path_2_zeros_needed = 10

    path_1_zero_bytes = path_1_zeros_needed // 8
    path_1_zero_bits_shift = 8 - (path_1_zeros_needed % 8)

    path_2_zero_bytes = path_2_zeros_needed // 8
    path_2_zero_bits_shift = 8 - (path_2_zeros_needed % 8)

    rules = (Rule(LinkExtractor(), callback='parse_item', follow=True, process_links='link_filtering'),)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.crawlerProcess = kwargs['crawlerProcess']
        self.spider_i = kwargs['spider_i']
        self.bw = kwargs['bw']
        self.min_path_len = kwargs['min_path_len']
        self.min_index_size = kwargs['min_index_size']

        self.mongo_client = MongoClient(self.mongo_uri)
        self.mongo_db = self.mongo_client[self.mongo_db]
        self.current_webcoin_lock = threading.RLock()
        self.current_webcoin = (0,1)
        self.spider_ip_bytes = socket.gethostname()
        self.spider_ip_str = '.'.join([str(x) for x in self.spider_ip_bytes])
        self.item_count = 0
        self.scraped_items_lock = threading.RLock()
        self.linkExtractor = LinkExtractor()
        self.urls_per_webcoin = []
        self.bodys_per_webcoin = []
        self.outbound_links_per = []

    def start_requests(self):
        start_options = list(self.mongo_db.start_urls.find())
        shuffle(start_options)
        url_dict = self.mongo_db.start_urls.find_one_and_delete(start_options[0])
        if url_dict:
            meta = {
                'current_webcoin' : self.current_webcoin,
                'path_1_urls' : [url_dict['url']],
                'path_2_urls' : [],
                'spider_i' :self.spider_i,
                'parent_hash_val' : bytes(),
                'path_1_hash_val' : bytes(),
                'path_2_hash_val' : bytes(),
            }
            start_url_request = Request(url_dict['url'], meta=meta)

            return [start_url_request]
        else:
            logging.error("Spider %s did not find any first URL to start, so it quits" % self.spider_i)
            return []

    def parse_start_url(self, response):
        return self.parse_item(response)

    def parse_item(self, response):
        # drop responses belonging to previous webcoin or path
        # if False:

        item = URLItem()
        item['url'] = response.url
        # item['url_id'] = self.double_hash(str.encode(response.url))
        # item['title'] = response.title
        # item['text_body'] = '\n'.join(response.xpath('//text()').extract())
        item['body'] = response.body
        # item['num_links'] = len(self.linkExtractor.extract_links(response))
        item['spider_i'] = self.spider_i
        item['bw'] = self.bw
        item['spider_ip'] = self.spider_ip_str
        item['current_webcoin'] = response.meta['current_webcoin']
        item['is_puzzle_solution'] = response.is_puzzle_solution

        item['path_1_urls'] = response.meta['path_1_urls']
        item['path_2_urls'] = response.meta['path_2_urls']
        item['path_1_hash_val'] = response.meta['path_1_hash_val']
        item['path_2_hash_val'] = response.meta['path_2_hash_val']
        if item['is_puzzle_solution']:
            if item['current_webcoin'][1] == 1:
                item['path_1_hash_val'] = response.hash_val
            else:
                item['path_2_hash_val'] = response.hash_val

        with self.current_webcoin_lock:
            item['ignore'] = not (response.meta['current_webcoin'] == self.current_webcoin or response.is_puzzle_solution)

        if (self.item_count % 100) == 0:
            logging.info('items_parsed: %s' % self.item_count)
        self.item_count += 1

        return item


    # Overriding CrawlSpider's method in order to check hashing of returned response.
    def _parse_response(self, response, callback, cb_kwargs, follow=True):
        try:
            # drop responses belonging to previous webcoin or path
            if response.meta.get('current_webcoin', None) != self.current_webcoin:
                return []
        except:
            print('what just happened?')
            return []


        # Check hash_val and if webpage is a solution to current puzzle
        is_puzzle_solution, response_hash_val = self.is_puzzle_solution(response)
        response.is_puzzle_solution = is_puzzle_solution
        response.hash_val = response_hash_val
        if is_puzzle_solution:
            # verify that no other winning webpage is found in parallel somehow
            with self.current_webcoin_lock:
                # verify that the winning URL is not of a path already completed
                if response.meta['current_webcoin'] == self.current_webcoin:
                    logging.info('TRUE Positive! \ncurrent_webcoin: %s \nitem_count: %s \nurl: %s \npath_1: %s \npath_2: %s' % (response.meta['current_webcoin'], self.item_count, response.url, response.meta['path_1_urls'], response.meta['path_2_urls']))
                    if self.current_webcoin[1] == 1:
                        self.current_webcoin = (self.current_webcoin[0], 2)
                    else:
                        self.current_webcoin = (self.current_webcoin[0] + 1, 1)
                    # response.request.current_webcoin= self.current_webcoin

                # If this is a "stale" winning URL, belonging to a path already completed, mark is_head = False
                else:
                    logging.info('FALSE Positive! \ncurrent_webcoin: %s \nitem_count: %s \nurl: %s \npath_1: %s \npath_2: %s' % (response.meta['current_webcoin'], self.item_count, response.url, response.meta['path_1_urls'], response.meta['path_2_urls']))
                    response.is_puzzle_solution = False



        # callback is parse_item() which creates the URLItem.
        # That means we need to check is_path_head() and hash_val before this point.
        if callback:
            cb_res = callback(response, **cb_kwargs) or ()
            cb_res = self.process_results(response, cb_res)
            for requests_or_item in iterate_spider_output(cb_res):
                yield requests_or_item

        if follow and self._follow_links:
            for request_or_item in self._requests_to_follow(response):
                # Pass parameters from parent_response to child_request
                self.adjust_child_request(request_or_item, response)
                # Skip all children if they are irrelevant
                if request_or_item.meta['current_webcoin'] != self.current_webcoin:
                    # return [request_or_item]
                    # continue
                    return []
                yield request_or_item

    def adjust_child_request(self, request_or_item, response):
        request_or_item.meta['spider_i'] = response.meta['spider_i']

        if response.is_puzzle_solution:
            request_or_item.meta['parent_hash_val'] = bytes()
            if response.meta['current_webcoin'][1] == 1:
                request_or_item.meta['current_webcoin'] = (response.meta['current_webcoin'][0], 2)
                request_or_item.meta['path_1_urls'] = response.meta['path_1_urls']
                request_or_item.meta['path_2_urls'] = [request_or_item.url]
                request_or_item.meta['path_1_hash_val'] = response.hash_val
                request_or_item.meta['path_2_hash_val'] = bytes()
            else:
                request_or_item.meta['current_webcoin'] = (response.meta['current_webcoin'][0] + 1, 1)
                request_or_item.meta['path_1_urls'] = [request_or_item.url]
                request_or_item.meta['path_2_urls'] = []
                request_or_item.meta['path_1_hash_val'] = bytes()
                request_or_item.meta['path_2_hash_val'] = bytes()
        else:
            request_or_item.meta['current_webcoin'] = response.meta['current_webcoin']
            request_or_item.meta['parent_hash_val'] = response.hash_val
            request_or_item.meta['path_1_hash_val'] = response.meta['path_1_hash_val']
            request_or_item.meta['path_2_hash_val'] = response.meta['path_2_hash_val']

            if response.meta['current_webcoin'][1] == 1:
                request_or_item.meta['path_1_urls'] = response.meta['path_1_urls'] + [request_or_item.url]
                request_or_item.meta['path_2_urls'] = []
            else:
                request_or_item.meta['path_1_urls'] = response.meta['path_1_urls']
                request_or_item.meta['path_2_urls'] = response.meta['path_2_urls'] + [request_or_item.url]


    def double_hash(self, to_hash):
        return sha256(sha256(to_hash).digest()).digest()

    def check_path(self, hash_val, path_type):
        # print('hash_val: %s '%[bin(x)[2:] for x in hash_val[:2]])
        if path_type == 1:
            for byte_i in range(self.path_1_zero_bytes):
                if hash_val[byte_i] != 0:
                    return False
            if (hash_val[self.path_1_zero_bytes] >> self.path_1_zero_bits_shift) > 0:
                return False
            return True

        else:
            for byte_i in range(self.path_2_zero_bytes):
                if hash_val[byte_i] != 0:
                    return False
            if (hash_val[self.path_2_zero_bytes] >> self.path_2_zero_bits_shift) > 0:
                return False
            return True

    def is_puzzle_solution(self, response):

        to_hash = response.body + response.meta['parent_hash_val']

        if response.meta['current_webcoin'][1] == 1:
            hash_val = self.double_hash(to_hash)
        else:
            hash_val = self.double_hash(to_hash + self.spider_ip_bytes)

        if len(response.meta['path_1_urls']) < self.min_path_len or len(self.urls_per_webcoin) < self.min_index_size:
            is_head = False
        else:
            is_head = self.check_path(hash_val, response.meta['current_webcoin'][1])
        return is_head, hash_val