import gzip
import lzma
import pickle
import string
import threading
from collections import defaultdict
from datetime import datetime, date
import os
from multiprocessing.pool import ThreadPool
from random import sample

from multiprocessing import Queue

import sys

import zlib

from math import ceil
from time import sleep

from pymongo import MongoClient
from hashlib import sha256

from scrapy.exceptions import DropItem
from scrapy_projects.itzy_bitzy_project.spiders.broad_crawl_spiders import BaseBitsySpider

# Define your item pipelines here
#
# Don't forget to add your pipeline to the ITEM_PIPELINES setting
# See: http://doc.scrapy.org/en/latest/topics/item-pipeline.html


class FilterAndStatsPipeline(object):
    def __init__(self):
        self.mongo_client = MongoClient(BaseBitsySpider.mongo_uri)
        self.mongo_db = self.mongo_client[BaseBitsySpider.mongo_db]

    def process_item(self, item, spider):
        if not item['last_item']:
            raise DropItem()
        else:
            stats = spider.crawler.stats.get_stats()
            stats['run_time_seconds'] = (item['finish_time'] - stats['webcoin_start_time']).total_seconds()
            stats['spider_ip'] = spider.spider_ip
            stats['bw'] = spider.bw
            stats['index_size'] = spider.index_size
            stats['start_urls_tier'] = spider.start_urls_tier

            crawl_stats = {}
            for key, value in stats.items():
                crawl_stats[key.replace('.', ':')] = value

            item['crawl_stats'] = crawl_stats

            return item

    def open_spider(self, spider):
        stats = spider.crawler.stats.get_stats()
        stats['webcoin_start_time'] = datetime.now()


class AddStartURLsPipeline(object):
    """
    The start_urls in mongoDB has 2 hierarchies:
    start_urls_tier:    each tier is created based on the tiers below it.
    seq_in_tier:        0 for the first in the tier, 1 for the 2nd, etc.
    We DO NOT consider the bw nor the index size for this purpose.
    """

    max_urls_per_tier = 10000

    def __init__(self):
        self.mongo_client = MongoClient(BaseBitsySpider.mongo_uri)
        self.mongo_db = self.mongo_client[BaseBitsySpider.mongo_db]

    def process_item(self, item, spider):
        next_tier = spider.start_urls_tier + 1

        start_urls_records = list(self.mongo_db.start_urls.find({'start_urls_tier': next_tier}).sort("seq_in_tier", -1))
        total_existing_urls = set(url for urls_batch in start_urls_records for url in urls_batch['urls'])
        if len(total_existing_urls) >= AddStartURLsPipeline.max_urls_per_tier:
            return item
        else:
            if start_urls_records:
                next_seq = start_urls_records[0]['seq_in_tier'] + 1
            else:
                next_seq = 0

            start_urls_dict = {
                'start_urls_tier': next_tier,
                'seq_in_tier': next_seq,
                'urls': spider.collected_urls,
                'spider_ip': spider.spider_ip,
                'bw': spider.bw,
                'urls_per_index': spider.index_size,
            }

        self.mongo_db.start_urls.insert(start_urls_dict)
        return item


def verify_decomprss_depickle(compressed_index, simplified_index, compressed_words_dict, all_words_dict):
    depickled_index = pickle.loads(lzma.decompress(compressed_index))
    depickled_words_dict = pickle.loads(lzma.decompress(compressed_words_dict))

    if depickled_words_dict != all_words_dict:
        print('Mismatch in pre/post pickling and compression of all_words_dict!')
        return False
    elif len(depickled_index) != len(simplified_index):
        print('Length mismatch in pre/post index pickling and compression')
        return False
    else:
        for page_i, page in enumerate(depickled_index):
            if page != simplified_index[page_i]:
                print('Mismatch in pre/post pickling and compression of page_i: %s in index' % page_i)
                return False
    return True


class CreateIndexPipeline(object):
    exclude_punctuation = set(string.punctuation)

    def __init__(self):
        self.mongo_client = MongoClient(BaseBitsySpider.mongo_uri)
        self.mongo_db = self.mongo_client[BaseBitsySpider.mongo_db]
        self.translator = str.maketrans({key: ' ' for key in string.punctuation + '1234567890'})

    def process_item(self, item, spider):
        word_count = []
        all_words_dict = {}
        all_words_dict_lock = threading.Lock()
        bodys_input = [(i, body, all_words_dict, all_words_dict_lock) for (i, body) in enumerate(spider.collected_bodys)]
        pool = ThreadPool(50)

        time0 = datetime.now()
        all_pages_words_iter = pool.imap_unordered(self.extract_words, bodys_input)
        sorted_all_pages_words = sorted([x for x in all_pages_words_iter], key=lambda x: x[0])
        simplified_index = [x[1] for x in sorted_all_pages_words]
        time1 = datetime.now()
        pool.terminate()

        # for i, body in enumerate(spider.collected_bodys):
        #     page_words = str.split(str(body).translate(self.translator).lower())
        #     word_count.append(len(page_words))
        #
        #     page_words_IDs = []
        #     for word in page_words:
        #         if word not in all_words_dict:
        #             all_words_dict[word] = len(all_words_dict)
        #         page_words_IDs.append(all_words_dict[word])
        #     simplified_index.append(page_words_IDs)

        print('-----------------------------------------------------------------------------')
        print('time to create simplified index: %s' % (time1 - time0).total_seconds())
        print('-----------------------------------------------------------------------------')

        pre_compression_time = datetime.now()
        pickled_index = pickle.dumps(simplified_index, protocol=-1)
        compressed_index = lzma.compress(pickled_index)
        post_compression_time = datetime.now()
        compressed_index_size_KB = ceil(sys.getsizeof(compressed_index) / 2**10)
        print('time to pickle+lzma index: %s seconds' % (post_compression_time - pre_compression_time).total_seconds())
        print('compressed_index_size_KB: %s' % compressed_index_size_KB)
        print('-----------------------------------------------------------------------------')


        pre_compression_time = datetime.now()
        pickled_words_dict = pickle.dumps(all_words_dict, protocol=-1)
        compressed_words_dict = lzma.compress(pickled_words_dict)
        post_compression_time = datetime.now()
        compressed_words_dict_size_KB = ceil(sys.getsizeof(compressed_words_dict) / 2**10)
        print('time to pickle+lzma words_dict: %s seconds' % (post_compression_time - pre_compression_time).total_seconds())
        print('compressed_words_dict_size_KB: %s' % compressed_words_dict_size_KB)
        print('-----------------------------------------------------------------------------')

        compression_verified = verify_decomprss_depickle(compressed_index, simplified_index, compressed_words_dict, all_words_dict)
        print('compression_verified: %s' % compression_verified)
        if not compression_verified:
            raise Exception('Shit!')
        print('-----------------------------------------------------------------------------')

        crawl_results_to_mongo = {
            'spider_ip': spider.spider_ip,
            'bw': spider.bw,
            'index_size': spider.index_size,
            'compressed_index_size_KB': compressed_index_size_KB,
            'compressed_words_dict_size_KB': compressed_words_dict_size_KB,
            'word_count': word_count,
            'urls': spider.collected_urls,
            'compressed_index': compressed_index,
            'compressed_words_dict': compressed_words_dict,
            'stats': item['crawl_stats'],
        }
        self.mongo_db.indexes_stats.insert(crawl_results_to_mongo)

        print('stats updated to mongo!')

        # bodys_enum = [(i, body) ]
        #
        #
        # all_pages_words_iter = self.pool.imap_unordered(self.extract_words, bodys_enum)
        # all_pages_words = [x for x in all_pages_words_iter]
        #
        #
        # real_index = create_compressable_index_function()
        # locks_dict = {}
        # master_lock = threading.Lock()
        # indexing_input = [(page, real_index, all_words_dict, locks_dict, master_lock) for page in all_pages_words]
        # added_pages_iter = self.pool.imap_unordered(self.add_to_index, indexing_input)
        # simplified_index = sorted([x for x in added_pages_iter], key=lambda x: x[0])
        # simplified_index = [x[0] for x in simplified_index]
        # time2 = datetime.now()
        # print('time to index words: %s' % (time2 - time1).total_seconds())
        #
        # for index in [real_index, simplified_index]:
        #     print('-----------------------------------------------------------------------------')
        #     pre_pickle_time = datetime.now()
        #     pickled = pickle.dumps(index, protocol=-1)
        #     post_pickle_time = datetime.now()
        #     print('time to pickle: %s' % (post_pickle_time - pre_pickle_time).total_seconds())
        #     print('pickle_size: %s MB' % (sys.getsizeof(pickled) / 2**20))
        #
        #     pre_gzip_time = datetime.now()
        #     gzipped_index = gzip.compress(pickled)
        #     post_gzip_time = datetime.now()
        #     print('time to gzip: %s' % (post_gzip_time - pre_gzip_time).total_seconds())
        #     print('gzip_size: %s MB' % (sys.getsizeof(gzipped_index) / 2**20))
        #
        #
        #     pre_lzma_time = datetime.now()
        #     # lzma_index = zlib.compress(pickled_index, level=9)
        #     lzma_index = lzma.compress(pickled)
        #     post_lzma_time = datetime.now()
        #     print('time to lzma: %s' % (post_lzma_time - pre_lzma_time).total_seconds())
        #     print('lzma_size: %s MB' % (sys.getsizeof(lzma_index) / 2**20))

        return item

    def add_to_index(self, indexing_input):
        (page, index, all_words_dict, locks_dict, master_lock) = indexing_input
        words_list = []
        for word_i, word in enumerate(page[1]):
            if word not in all_words_dict:
                with master_lock:
                    if word not in all_words_dict:
                        all_words_dict[word] = len(all_words_dict)
                        locks_dict[word] = threading.Lock()

            with locks_dict[word]:
                index[all_words_dict[word]].append(([page[0]], word_i))
            words_list.append(all_words_dict[word])
        return (page[0], words_list)

    def extract_words(self, body_input):
        page_i, body, all_words_dict, all_words_dict_lock = body_input
        page_words_IDs = []

        page_words = str.split(str(body).translate(self.translator).lower())
        for word in page_words:
            if word not in all_words_dict:
                with all_words_dict_lock:
                    if word not in all_words_dict:
                        all_words_dict[word] = len(all_words_dict)

            page_words_IDs.append(all_words_dict[word])
        return page_i, page_words_IDs

def create_compressable_index_function():
    return defaultdict(list)
#     return defaultdict(compressable_index_function)
#
# def compressable_index_function():
#     return defaultdict(list)

class KeepSpiderAlivePipeline(object):

    def __init__(self):
        self.mongo_client = MongoClient(BaseBitsySpider.mongo_uri)
        self.mongo_db = self.mongo_client[BaseBitsySpider.mongo_db]

    def close_spider(self, spider):
        print('------------------------------------------')
        print('------------------------------------------')
        success = list(self.mongo_db.crawl_success_rate.find({'crawl_successful': True}))
        fail = list(self.mongo_db.crawl_success_rate.find({'crawl_successful': False}))
        if len(spider.collected_urls) < spider.index_size:
            print('spider failed! current_size: %s index_size: %s' %
              (len(spider.collected_urls), spider.index_size))
        else:
            print('spider success!')
        print('Success ratio 1:%s' %  (len(fail) / max(0.0001,len(success))))
        print('------------------------------------------')
        print('------------------------------------------')

        sleep(1)

        success_rate_to_mongo = {
            'spider_ip': spider.spider_ip,
            'bw': spider.bw,
            'crawl_successful': len(spider.collected_urls) >= spider.index_size,
            'index_size': spider.index_size,
            'current_size': len(spider.collected_urls),
            'urls': spider.collected_urls,
        }
        if len(spider.collected_urls) >= 10:
            self.mongo_db.crawl_success_rate.insert(success_rate_to_mongo)

        spider.crawlerProcess.crawl('backpointerwebcoincrawler', crawlerProcess=spider.crawlerProcess,
                                    bw=spider.bw, start_urls_tier=spider.start_urls_tier, index_sizes=spider.index_sizes)